import pandas as pd
import joblib
from datetime import datetime
from vulnscanpy.lib import preprocess, config_models, learning
from sklearn.model_selection import GridSearchCV

from sklearn.pipeline import Pipeline

from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_auc_score


def analysis(args, dataset, DEBUG):
    """ Put all step together and train models. """
    X_train, X_test, y_train, y_test, validator, steps, params, n_class, weight = preprocess(dataset, args.binary, args.multiclass, DEBUG)

    # scoring = 'roc_auc'
    # if args.multiclass:
    scoring = [ 'f1_weighted', 'balanced_accuracy' ]
    list_of_model = config_models(n_class, weight)
    scoring_text = "Used {} as scoring method\n\n".format(scoring)


    if DEBUG > 1:
        print([ model.get_model() for model in list_of_model])

    for score in scoring:
        for model in list_of_model:
            error_message = ""
            try:

                new_steps, new_params = learning(model, steps, params)
                njob = -1
                pipeline = Pipeline(new_steps)
                grid = GridSearchCV(pipeline, scoring=score, cv=validator, param_grid=new_params, n_jobs=njob, verbose=DEBUG*2) # refit='balanced',
                grid.fit(X_train, y_train)

                weight_text = "Weight: " + str(weight)
                if DEBUG > 0: print(weight_text)

                # Summarize results
                best_training_score = "Best training score for {}: {}\n\tUsing {}\n\n".format(model.get_tag_name(), grid.best_score_, grid.best_params_)
                if DEBUG > 0: print(best_training_score)

                best_estimator = grid.best_estimator_

                ordered_features = pd.DataFrame(dict(feature_names=X_train.columns, scores=best_estimator.steps[1][1].scores_)).sort_values('scores', ascending=False)
                ordered_features_text = "Ordered feature by select k best:\n" + str(ordered_features) + "\n\n"
                if DEBUG > 0:
                    print(ordered_features_text)

                tscore = grid.score(X_test, y_test)
                test_score = "Final score on test set: {}\n\n".format(tscore)
                if DEBUG > 0: print(test_score)

                y_pred = grid.predict(X_test)

                report = classification_report(y_test, y_pred)
                report_text = "Classification Report:\n" + str(report) + "\n\n"
                if DEBUG > 0: print(report_text)
                
                confusion = confusion_matrix(y_test, y_pred)
                confusion_text = "Confusion Matrix:\n" + str(confusion) + "\n\n"
                if DEBUG > 0: print(confusion_text)

            except Exception as e:
                print("Unhandle error!")
                print(e)
                error_message = str(e)
            
            model_name = "trained_" + model.get_tag_name() + "_"
            model_name += "multi_" if args.multiclass else "binary_"
            model_name += datetime.now().strftime("%y%m%d")

            filename = model_name + '_' + score + '.model'
            joblib.dump(grid, filename)

            with open(model_name + '_' + score + '.summary', "w") as f:
                f.write(str(weight_text))
                f.write(str(error_message) + "\n")
                f.write(str(scoring_text))
                f.write(str(best_training_score))
                f.write(str(ordered_features_text))
                f.write(str(test_score))
                f.write(str(report_text))
                f.write(str(confusion_text))
