import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split, PredefinedSplit
# Preprocessing method used
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.feature_selection import SelectKBest, mutual_info_classif, chi2

def _anonymize_data(X):
    """ Drop ip and port from the dataset to anonymize the data """
    return X.drop(['c_client_ip', 'c_client_port', 's_server_ip', 's_server_port', 'c_internal', 's_internal'], axis=1)

def _remove_lab_from_dataset(X):
    """ Drop time oriented data due to labs testing area """
    return X.drop(['s_ttl_max', 'c_ttl_min', 'c_ttl_max', 's_ttl_min', 'first_time_abs', 'last_time_abs', 'rtt_c_avg', 'rtt_c_max', 'rtt_c_std', 'rtt_s_max'], axis=1)

def _divide_dataset(X, y):
    """ Divide data into Training, Validation and Test set """
    X_train, X_other, y_train, y_other = train_test_split(
        X, y, train_size=0.7, random_state=80085)
    X_validation, X_test, y_validation, y_test = train_test_split(
        X_other, y_other, test_size=0.5, random_state=80085)

    # Build new train set
    X_tmp = pd.concat([X_train, X_validation])
    y_tmp = pd.concat([y_train, y_validation])

    # Index for validation set
    validation_index = [-1 if x in X_train.index else 0 for x in X_tmp.index]

    # Validator with validation set
    validator = PredefinedSplit(validation_index)

    # Return (X_train+X_validation), X_test, (y_train+y_validation), y_test, validator
    return X_tmp, X_test, y_tmp, y_test, validator

def get_class_weight(y, n_class):
    weight = []
    for i in range(n_class):
        h = y[y[:] == i]
        weight.append( 1.0 / len(h) )        
    return weight

# def get_class_weight(y):
#     from sklearn.utils import class_weight
#     weight = class_weight.compute_class_weight('balanced', np.unique(y), y)
#     return weight

def preprocess(dataset, binary=None, multiclass=None, DEBUG=0):
    """ All action releted to the preprocess of the data. Return splitted data, step of pipeline and params for GridsearchCV. """
    if binary and multiclass: raise Exception("Not yet supported")
    X = y = None
    n_class = 2

    from sklearn.preprocessing import LabelEncoder
    le = LabelEncoder()
    
    dataset = _anonymize_data(dataset).drop_duplicates()
    dataset = _remove_lab_from_dataset(dataset)

    if DEBUG > 2: print(dataset.corr())

    if binary:
        if 'class' in dataset: dataset = dataset.drop('class', axis=1) 
        X = dataset.drop(binary, axis=1).astype(np.float32)
        y = dataset[binary]
    elif multiclass:
        if 'evil' in dataset: dataset = dataset.drop('evil', axis=1) 
        X = dataset.drop(multiclass, axis=1).astype(np.float32)

        distribution = dataset[multiclass].value_counts()
        if DEBUG > 0: print(f"Total Distribution:\n{distribution}")

        y = pd.Series(le.fit_transform(dataset[multiclass]))
        n_class = le.classes_.shape[0]
     
    y = y.astype(np.int64)

    if DEBUG > 0:
        print("X shape: ", X.shape)
        print("y shape: ", y.shape)

    # Divide dataset into Training and Test set plus validator for optimization of hyperparameter
    X_train, X_test, y_train, y_test, validator = _divide_dataset(X, y)

    y_train_label = pd.Series(le.inverse_transform(y_train))
    y_test_label = pd.Series(le.inverse_transform(y_test))
    train_distribution = y_train_label.value_counts()
    test_distribution = y_test_label.value_counts()

    if DEBUG > 0:
        print(f"Train Distribution:\n{train_distribution}")
        print(f"Test Distribution:\n{test_distribution}")


    weight = get_class_weight(y_train, n_class)
    if DEBUG > 0: print(f"Weight of the class: {weight}")

    # weight_all = weight[y_train]
    
    # Standardization
    base_scaler = ('scaler', StandardScaler())
    scalers = [StandardScaler(), MinMaxScaler()]
    
    # Feature Selection
    selector = ('kbest', SelectKBest())
    k_selection = [ 20 ]

    params = [
        {
            "scaler": scalers,
            "kbest__score_func": [ mutual_info_classif ],
            "kbest__k": k_selection
        },
        {
            "scaler": [MinMaxScaler()],
            "kbest__score_func": [chi2],
            "kbest__k": k_selection
        }
    ]

    steps = [ base_scaler, selector ]

    return X_train, X_test, y_train, y_test, validator, steps, params, n_class, weight #_all
