from torch.utils.data import Dataset, DataLoader, WeightedRandomSampler
from vulnscanpy.models.Dummy import DummyModel
from torch import nn
from skorch import NeuralNetClassifier
import skorch
import torch

class ClassifierModule(nn.Module):
    def __init__(self, hidden_size=[1,2], input_dim=80, output_dim=2, nonlin='sigmoid', 
                        norm=None, last_stage=None, dropout=None):
        super(ClassifierModule, self).__init__()

        # Non-linear function
        if "sigmoid" in nonlin:
            self.nonlin = nn.Sigmoid()
        elif "relu" in nonlin:
            self.nonlin = nn.ReLU()
        
        # If set use dropout
        self.dropout = nn.Dropout(dropout) if dropout else None
        
        if "soft" in last_stage:
            # Softmax
            self.last = nn.Softmax(dim=-1)
        elif "sigmoid" in last_stage:
            self.last = nn.Sigmoid()
        
        self.input = nn.Linear(input_dim, hidden_size[0])

        # Build hidden layers
        self.hidden = nn.ModuleList()

        if norm:
            self.hidden.append(nn.BatchNorm1d(hidden_size[0]))
        self.hidden.append(self.nonlin)

        for k in range(len(hidden_size)-1):
            self.hidden.append(nn.Linear(hidden_size[k], hidden_size[k+1]))
            if norm:
                self.hidden.append(nn.BatchNorm1d(hidden_size[k+1]))
            self.hidden.append(self.nonlin)
            if dropout:
                self.hidden.append(self.dropout)

        # Output Layer
        self.output = nn.Linear(hidden_size[-1], output_dim)

    def forward(self, X, **kwargs):
        X = self.input(X)

        for layer in self.hidden:
            X = layer(X)

        if self.last:
            output = self.last(self.output(X))
        else:
            output = self.output(X)

        return output


class InputShapeSetter(skorch.callbacks.Callback):
    def on_train_begin(self, net, X, y):
        net.set_params(module__input_dim=X.shape[1])

class NN(DummyModel):
    def __init__(self,  tag_name='nn', weight= None, **kwargs):
        model = None
        if len(weight) > 0:
            class_weights = torch.Tensor(weight)

            model = NeuralNetClassifier(ClassifierModule, criterion=torch.nn.CrossEntropyLoss, criterion__weight=class_weights, 
                                        optimizer=torch.optim.Adam, callbacks=[InputShapeSetter()], **kwargs)
        else:
            model = NeuralNetClassifier(ClassifierModule, callbacks=[InputShapeSetter()], **kwargs)
        super().__init__(tag_name, model)

    def my_params(self):
        # Learning rate
        lr = [0.0007, 0.05, 0.1, 0.2]

        # Hidden layers
        hiddens = [
            # [ 5, 3  ],            # 1 hidden layer
            # [ 10, 5 ],            # 1 hidden layer
            # [ 20, 10, 5  ],         # 2 hidden layer
            # [ 10, 20, 5 ],          # 2 hidden layer
            # [ 15, 30, 10, 5 ],      # 3 hidden layer
            # [ 15, 30, 20, 10, 5 ],  # 4 hidden layer
            [ 512, 128, 64 ],       # Try greater number
            [ 512, 256, 128, 64 ]
        ]

        # Dropout
        dropout = [None, 0.2, 0.5]
        
        # Non-lienar
        nonlin = [ 'relu' ]

        # BatchNorm
        norm = [None, True]

        # Last stage function
        last_s = ['soft', 'sigmoid']

        return [{
            'lr': lr,
            'module__dropout': dropout,
            'module__nonlin': nonlin,
            'module__hidden_size': hiddens,
            'module__norm': norm,
            'module__last_stage': last_s
        }]
